import numpy as np
import pandas as pd
import networkx as nx
import os
import json
import pickle
import bz2
import _pickle as cPickle
from os.path import join
from random import shuffle
import time

from scseirx import construct_school_network as csn

def get_agent(model, ID):
    for a in model.schedule.agents:
        if a.ID == ID:
            return a

def test_infection(a):
    if a.infectious or a.recovered or a.exposed:
        return 1
    else:
        return 0
    
def count_infected(model, agent_type):
    infected_agents = np.asarray([test_infection(a) for a in model.schedule.agents \
                         if a.type == agent_type]).sum()
    
    return infected_agents

def count_infection_endpoints(model):
    endpoints = [a for a in model.schedule.agents if a.recovered and \
         a.transmissions == 0]

    return len(endpoints)

def count_typed_transmissions(model, source_type, target_type):
    type_dict = {'t':'teacher', 's':'student', 'f':'family_member', \
        'r':'resident', 'e':'employee'}
    sources = [a for a in model.schedule.agents if a.type == source_type]
    transmissions = 0
    for source in sources:
        for target, step in source.transmission_targets.items():
            if type_dict[target[0]] == target_type:
                transmissions += 1
    return transmissions

    
def calculate_R0(model, agent_types):
    transmissions = [a.transmissions for a in model.schedule.agents]
    infected = [test_infection(a) for a in model.schedule.agents]
    IDs = [a.ID for a in model.schedule.agents]
    types = [a.type for a in model.schedule.agents]
    df = pd.DataFrame({'ID':IDs,
                       'type':types,
                       'was_infected':infected,
                       'transmissions':transmissions})
    df = df[df['was_infected'] == 1]
    overall_R0 = df['transmissions'].mean()


    R0 = {}
    for agent_type in agent_types:
        agent_R0 = df[df['type'] == agent_type]['transmissions'].mean()
        R0.update({agent_type:agent_R0})
    
    return R0

def calculate_finite_size_R0(model):
    df = pd.DataFrame(columns=['ID', 'agent_type', 't', 'target'])
    for a in model.schedule.agents:
        if a.transmissions > 0:
            for target in a.transmission_targets.keys():
                df = df.append({'ID':a.ID, 'agent_type':a.type,
                    't':a.transmission_targets[target], 'target':target},
                            ignore_index=True)
                
    # find first transmission(s)
    # NOTE: while it is very unlikely that two first transmissions occurred
    # in the same timestep, we have to account for the possibility nevertheless
    first_transmitters = df[df['t'] == df['t'].min()]['ID'].values
    N_transmissions = []
    for ft in first_transmitters:
        N_transmissions.append(len(df[df['ID'] == ft]))
    
    if len(N_transmissions) > 0:
        mean = np.mean(N_transmissions)
    else:
        mean = 0

    return mean, df
    

def count_infected_by_age(model, age_brackets):
	age_counts = {}
	for ab in age_brackets:
	    lower = int(ab.split('-')[0])
	    upper = int(ab.split('-')[1])
	    
	    infected = len([a for a in model.schedule.agents if a.type == 'student' and \
	               a.recovered == True and a.age >= lower and a.age <= upper])
	    
	    age_counts[ab] = infected

	return age_counts


def get_transmission_network(model):
    transmissions = []
    for a in model.schedule.agents:
        if a.transmissions > 0:
            for target in a.transmission_targets.keys():
                transmissions.append((a.ID, target))
                
    G = nx.Graph()
    G.add_edges_from(transmissions)
                
    return G

def get_statistics(df, col):
    if len(df) == 0:
        return {
            '{}_mean'.format(col):np.nan,
            '{}_median'.format(col):np.nan,
            '{}_0.025'.format(col):np.nan,
            '{}_0.10'.format(col):np.nan,
            '{}_0.25'.format(col):np.nan,
            '{}_0.75'.format(col):np.nan,
            '{}_0.90'.format(col):np.nan,
            '{}_0.975'.format(col):np.nan,
            '{}_std'.format(col):np.nan
        }
    else:
        return {
            '{}_mean'.format(col):df[col].mean(),
            '{}_median'.format(col):df[col].median(),
            '{}_0.025'.format(col):df[col].quantile(0.025),
            '{}_0.10'.format(col):df[col].quantile(0.10),
            '{}_0.25'.format(col):df[col].quantile(0.25),
            '{}_0.75'.format(col):df[col].quantile(0.75),
            '{}_0.90'.format(col):df[col].quantile(0.90),
            '{}_0.975'.format(col):df[col].quantile(0.975),
            '{}_std'.format(col):df[col].std(),
        }

def get_agent_states(model, tm_events):
    if type(tm_events) == type(None):
        return None
    # all agent states in all simulation steps. Agent states include the
    # "infection state" ["susceptible", "exposed", "infectious", "recovered"]
    # as well as the "quarantine state" [True, False]
    state_data = model.datacollector.get_agent_vars_dataframe()
    # remove susceptible states: these are the majority of states and since we
    # want to reduce the amoung of data stored, we will assume all agents that
    # do not have an explicit "exposed", "infectious" or "recovered" state, 
    # are susceptible.
    state_data = state_data[state_data['infection_state'] != 'susceptible']

    # we only care about state changes here, that's why we only keep the first
    # new entry in the state data table for each (agent, infection_state,
    # quarantine_state) triple. We need to reset the index once before we can
    # apply the drop_duplicates() operation, since "step" and "AgentID" are in
    # the index
    state_data = state_data.reset_index()
    state_data = state_data.drop_duplicates(subset=['AgentID',\
        'infection_state', 'quarantine_state'])

    # cleanup of index and column names to match other data output formats
    state_data = state_data.rename(columns={'AgentID':'node_ID',
                                            'Step':'day'})
    state_data = state_data.reset_index(drop=True)

    # for the visualization we need more fine-grained information (hours) on when 
    # exactly a transmission happened. This information is already stored in the
    # transmission events table (tm_events) and we can take it from there and
    # add it to the state table. We set all state changes to hour = 0 by default
    # since most of them happen at the beginning of the day (becoming infectious,
    # changing quarantine state or recovering). 
    state_data['hour'] = 1
    exp = state_data[(state_data['infection_state'] == 'exposed')]\
        .sort_values(by='day')

    # we iterate over all state changes that correspond to an exposure, ignoring
    # the first agent, since that one is the index case, whose exposure happens
    # in hour 0. We only add the hour information for the transmission events. 
    # For these events, we also subtract one from the event "day", to account 
    # for the fact that agents have to appear "exposed" in the visualization 
    # from the point in time they had contact with an infectious agent onwards. 
    for ID in exp['node_ID'][1:]:
        ID_data = exp[exp['node_ID'] == ID]
        idx = ID_data.index[0]
        hour = tm_events[tm_events['target_ID'] == ID]['hour'].values[0]
        state_data.loc[idx, 'hour'] = hour
        state_data.loc[idx, 'day'] -= 1

    # re-order columns to a more sensible order and fix data types
    state_data['hour'] = state_data['hour'].astype(int)
    state_data = state_data[['day', 'hour', 'node_ID', 'infection_state',
                                'quarantine_state']]

    return state_data


def get_transmission_chain(model, school_type, teacher_schedule, student_schedule):

    max_hours = 9
    teaching_hours = csn.get_teaching_hours(school_type)
    daycare_hours = list(range(teaching_hours + 1, max_hours + 1))
    teaching_hours = list(range(1, teaching_hours + 1))

    weekday_map = {1:'monday', 2:'tuesday', 3:'wednesday', 4:'thursday',
                   5:'friday', 6:'saturday', 7:'sunday'}
    
    if 5 in daycare_hours:
        daycare_hours.remove(5)


    tm_events = pd.DataFrame(columns=['day', 'weekday', 'hour',
        'source_ID', 'source_type', 'target_ID', 'target_type'])

    for a in model.schedule.agents:
        if a.transmissions > 0:
            for target, step in a.transmission_targets.items():
                location = ''
                hour = np.nan
                weekday =  (step + model.weekday_offset) % 7 + 1
                G = model.weekday_connections[weekday]
                target = get_agent(model, target)
                n1 = a.ID
                n2 = target.ID
                tmp = [n1, n2]
                tmp.sort()
                n1, n2 = tmp
                key = n1 + n2 + 'd{}'.format(weekday)

                s_schedule = student_schedule.loc[weekday]
                t_schedule = teacher_schedule.loc[weekday]

                ## determine transmission locations and times
                # transmissions from students to other students, teachers or family
                # members
                if a.type == 'student':
                    student_class = G.nodes(data=True)[a.ID]['unit']
                    
                    if target.type == 'student':
                        # transmussions during daycare
                        if G[a.ID][target.ID][key]['link_type'] == 'student_student_daycare':
                            classroom = s_schedule.loc[a.ID]['hour_8']
                            location = 'class_{}'.format(int(classroom))
                            hour = np.random.choice(daycare_hours)
                        # transmission during morning teaching
                        elif G[a.ID][target.ID][key]['link_type'] in \
                            ['student_student_intra_class', 'student_student_table_neighbour']:
                            hour = np.random.choice(teaching_hours)
                            classroom = s_schedule.loc[a.ID]['hour_1']
                            location = 'class_{}'.format(int(classroom))  
                        elif G[a.ID][target.ID][key]['link_type'] == 'student_household':
                            hour = 10
                            location = 'home'                     
                        else:
                            print('unknown student <-> student link type ',\
                            G[a.ID][target.ID][key]['link_type'])
                        
                    # transmissions between students and teachers occur in the student's
                    # classroom at a time when the teacher is in that classroom
                    # according to the schedule
                    elif target.type == 'teacher':
                        # transmissions during daycare
                        if G[a.ID][target.ID][key]['link_type'] == 'daycare_supervision_teacher_student':
                            classroom = s_schedule.loc[a.ID]['hour_8']
                            location = 'class_{}'.format(int(classroom))
                            hour = np.random.choice(daycare_hours)
                        elif G[a.ID][target.ID][key]['link_type'] == 'teaching_teacher_student':
                            classroom = s_schedule.loc[a.ID]['hour_1']
                            location = 'class_{}'.format(int(classroom))
                            # get the hour in which the teacher is teaching in the given location
                            hour = int(t_schedule.loc[target.ID][t_schedule.loc[target.ID] == classroom]\
                                .index[0].split('_')[1])                          
                        else:
                            print('unknown student <-> teacher link type', \
                                G[a.ID][target.ID][key]['link_type'])
                    
                    # transmissions to family members occur at home after schoole
                    elif target.type == 'family_member':
                        location = 'home'
                        hour = 10
                        
                    else:
                        print('agent type not supported')

                # transmissions from teachers to other teachers or students
                elif a.type == 'teacher':
                    # transmissions from teachers to students occur in the student's
                    # classroom at a time when the teacher is in that classroom
                    # according to the schedule
                    if target.type == 'student':
                        # transmissions during daycare
                        if G[a.ID][target.ID][key]['link_type'] == 'daycare_supervision_teacher_student':
                            classroom = s_schedule.loc[target.ID]['hour_8']
                            location = 'class_{}'.format(int(classroom))
                            hour = np.random.choice(daycare_hours)
                        elif G[a.ID][target.ID][key]['link_type'] == 'teaching_teacher_student':
                            classroom = s_schedule.loc[target.ID]['hour_1']
                            location = 'class_{}'.format(int(classroom))
                            # get the hour in which the teacher is teaching in the given location
                            hour = int(t_schedule.loc[a.ID][t_schedule.loc[a.ID] == classroom]\
                                .index[0].split('_')[1])     
                        else:
                            print('unknown teacher <-> student link type', \
                                G[a.ID][target.ID][key]['link_type'])
                        
                    # transmissions between teachers occur during the lunch break
                    # in the faculty room
                    elif target.type == 'teacher':
                        location = 'faculty_room'
                        hour = 5
                        
                    elif target.type == 'family_member':
                        location = 'home'
                        hour = 10
                        
                    else:
                        print('agent type not supported')

                # transmissions from family members to other family members
                elif a.type == 'family_member':
                    if target.type == 'student':
                        location = 'home'
                        hour = 10
                        
                    elif target.type == 'teacher':
                        print('this should not happen!')
                        
                    # transmissions between family members occur at home after school
                    elif target.type == 'family_member':
                        location = 'home'
                        hour = 10
                        
                    else:
                        print('agent type not supported')

                else:
                    print('agent type not supported')
                
                assert not np.isnan(hour), 'schedule messup!'
                assert len(location) > 0, 'location messup!'
                tm_events = tm_events.append({
                    'day':step,
                    'weekday':weekday_map[weekday], 
                    'hour':hour,
                    'location':location,
                    'source_ID':a.ID,
                    'source_type':a.type,
                    'target_ID':target.ID,
                    'target_type':target.type},
                ignore_index=True)


    if len(tm_events) > 0:            
        tm_events['day'] = tm_events['day'].astype(int)
        tm_events = tm_events.sort_values(by=['day', 'hour']).reset_index(drop=True)
        tm_events = tm_events[['day', 'hour', 'location', 'source_ID', 'source_type',
                              'target_ID', 'target_type']]
        return tm_events
    else:
        return None

def get_ensemble_observables_school(model, run):
    R0, _ = calculate_finite_size_R0(model)
    N_school_agents = len([a for a in model.schedule.agents if \
        a.type == 'teacher' or a.type == 'student'])
    N_family_members = len([a for a in model.schedule.agents if a.type == 'family_member'])
    infected_students = count_infected(model, 'student')
    infected_teachers = count_infected(model, 'teacher')
    infected_family_members = count_infected(model, 'family_member')
    infected_agents = infected_students + infected_teachers + infected_family_members
    data = model.datacollector.get_model_vars_dataframe()
    N_diagnostic_tests = data['N_diagnostic_tests'].max()
    N_preventive_screening_tests = data['N_preventive_screening_tests'].max()
    transmissions = sum([a.transmissions for a in model.schedule.agents])
    infected_without_transmissions = count_infection_endpoints(model)
    student_student_transmissions = count_typed_transmissions(model, 'student', 'student')
    teacher_student_transmissions = count_typed_transmissions(model, 'teacher', 'student')
    student_teacher_transmissions = count_typed_transmissions(model, 'student', 'teacher')
    teacher_teacher_transmissions = count_typed_transmissions(model, 'teacher', 'teacher')
    student_family_member_transmissions = count_typed_transmissions(model, 'student', 'family_member')
    family_member_family_member_transmissions = count_typed_transmissions(model, 'family_member', 'family_member')
    quarantine_days_student = model.quarantine_counters['student']
    quarantine_days_teacher = model.quarantine_counters['teacher']
    quarantine_days_family_member = model.quarantine_counters['family_member']
    diagnostic_test_detected_infections_student = \
        data['diagnostic_test_detected_infections_student'].max()
    diagnostic_test_detected_infections_teacher = \
        data['diagnostic_test_detected_infections_teacher'].max()
    diagnostic_test_detected_infections_family_member = \
        data['diagnostic_test_detected_infections_family_member'].max()
    preventive_test_detected_infections_student = \
        data['preventive_test_detected_infections_student'].max()
    preventive_test_detected_infections_teacher = \
        data['preventive_test_detected_infections_teacher'].max()
    preventive_test_detected_infections_family_member = \
        data['preventive_test_detected_infections_family_member'].max()
    pending_test_infections = data['pending_test_infections'].max()
    undetected_infections = data['undetected_infections'].max()
    predetected_infections = data['predetected_infections'].max()
    duration = len(data)
    diagnostic_tests_per_day_per_agent = N_diagnostic_tests / duration / N_school_agents
    preventive_tests_per_day_per_agent = N_preventive_screening_tests / duration / N_school_agents
    tests_per_day_per_agent = (N_diagnostic_tests + N_preventive_screening_tests) / duration / N_school_agents

    row = {'run':run, 
          'R0':R0,
          'N_school_agents':N_school_agents,
          'N_family_members':N_family_members,
          'infected_students':infected_students,
          'infected_teachers':infected_teachers,
          'infected_family_members':infected_family_members,
          'infected_agents':infected_agents,
          'N_diagnostic_tests':N_diagnostic_tests,
          'N_preventive_tests':N_preventive_screening_tests,
          'transmissions':transmissions,
          'infected_without_transmissions':infected_without_transmissions,
          'student_student_transmissions':student_student_transmissions,
          'teacher_student_transmissions':teacher_student_transmissions,
          'student_teacher_transmissions':student_teacher_transmissions,
          'teacher_teacher_transmissions':teacher_teacher_transmissions,
          'student_family_member_transmissions':student_family_member_transmissions,
          'family_member_family_member_transmissions':family_member_family_member_transmissions,
          'quarantine_days_student':quarantine_days_student,
          'quarantine_days_teacher':quarantine_days_teacher,
          'quarantine_days_family_member':quarantine_days_family_member,
          'preventive_test_detected_infections_student':\
                preventive_test_detected_infections_student,
          'preventive_test_detected_infections_teacher':\
                preventive_test_detected_infections_teacher,
          'preventive_test_detected_infections_family_member':\
                preventive_test_detected_infections_family_member,
          'diagnostic_test_detected_infections_student':\
                diagnostic_test_detected_infections_student,
          'diagnostic_test_detected_infections_teacher':\
                diagnostic_test_detected_infections_teacher,
          'diagnostic_test_detected_infections_family_member':\
                diagnostic_test_detected_infections_family_member,
          'pending_test_infections':pending_test_infections,
          'undetected_infections':undetected_infections,
          'predetected_infections':predetected_infections,
          'duration':duration,
          'diagnostic_tests_per_day_per_agent':diagnostic_tests_per_day_per_agent,
          'preventive_tests_per_day_per_agent':preventive_tests_per_day_per_agent,
          'tests_per_day_per_agent':tests_per_day_per_agent}

    return row


def compress_pickle(fname, fpath, data):
    success = False
    while not success:
        try:
            with bz2.BZ2File(join(fpath, fname + '.pbz2'), 'w') as f: 
                cPickle.dump(data, f)
                success = True
        except OSError:
            time.sleep(0.5)
            print('re-trying to dump model file {} ...'.format(fname))
    return
    
    
def decompress_pickle(fname, fpath):
    data = bz2.BZ2File(join(fpath, fname), 'rb')
    data = cPickle.load(data)
    return data


def get_representative_run(N_infected, path):
    filenames = os.listdir(path)
    shuffle(filenames)
    medians = {int(f.split('_')[1]):int(f.split('_')[3].split('.')[0]) \
               for f in filenames}
    dist = np.inf
    closest_run = None
    
    for run, median in medians.items():
        curr_dist = np.abs(N_infected - median)
        if curr_dist < dist:
            closest_run = run
            dist = curr_dist
        if curr_dist == 0:
            break
            
    fname = 'run_{}_N_{}.pbz2'.format(closest_run, medians[closest_run])
    return decompress_pickle(fname, path)


def dump_JSON(path, school,
              test_type, index_case, screen_frequency_student, 
              screen_frequency_teacher, teacher_mask, student_mask, half_classes,
              ventilation_mod, node_list, teacher_schedule, student_schedule, 
              rep_transmission_events, state_data, start_weekday, duration,
              fname_addition='', friendship_contacts=False,
              class_size_reduction=False,
              m_efficiency_exhale=False, m_efficiency_inhale=False,
              s_test_rate=False, t_test_rate=False,
              trisk_mod = False):

    student_schedule = student_schedule.reset_index()
    teacher_schedule = teacher_schedule.reset_index()

    school_type = school['type']
    classes = school['classes']
    students = school['students']

    turnover, _, ttype = test_type.split('_')
    turnovers = {'same':0, 'one':1, 'two':2, 'three':3}
    turnover = turnovers[turnover]
    bool_dict = {True:'T', False:'F'}
    
    node_list = json.loads(node_list.to_json(orient='split'))
    del node_list['index']
    teacher_schedule = json.loads(teacher_schedule.to_json(orient='split'))
    del teacher_schedule['index']
    student_schedule = json.loads(student_schedule.to_json(orient='split'))
    del student_schedule['index']

    # can be empty, if there are no transmission events in the simulation
    try:
        rep_transmission_events = json.loads(rep_transmission_events\
            .to_json(orient='split'))
        del rep_transmission_events['index']
        state_data = json.loads(state_data.to_json(orient='split'))
        del state_data['index']
    except AttributeError:
        pass

    
    data = {'school_type':school_type,
            'classes':classes,
            'students':students,
            'test_type':ttype,
            'test_turnover':turnover,
            'indexcase':index_case,
            'screen_frequency_teacher':screen_frequency_teacher,
            'screen_frequency_student':screen_frequency_student,
            'teacher_mask':teacher_mask,
            'student_mask':student_mask,
            'half_classes':half_classes,
            'ventilation_mod':ventilation_mod,
            'node_list':node_list,
            'teacher_schedule':teacher_schedule,
            'student_schedule':student_schedule,
            'rep_trans_events':rep_transmission_events,
            'agent_states':state_data,
            'start_weekday':start_weekday,
            'duration':duration}


    fname = join(path, 'test-{}_'.format(ttype) + \
       'turnover-{}_index-{}_tf-{}_'
       .format(turnover, index_case[0], screen_frequency_teacher) +\
       'sf-{}_tmask-{}_smask-{}'\
       .format(screen_frequency_student, bool_dict[teacher_mask],\
        bool_dict[student_mask]))


    if friendship_contacts:
        fname = fname + '_fcontacts-{}'.format(friendship_contacts)
    if class_size_reduction:
        fname = fname + '_csizered-{}'.format(class_size_reduction)
    if m_efficiency_exhale and m_efficiency_inhale:
        fname = fname + '_meffinh-{}_meffexh-{}'\
            .format(m_efficiency_exhale, m_efficiency_inhale)
    if s_test_rate and t_test_rate:
        fname = fname + '_stestrate-{}_ttestrate-{}'\
        .format(s_test_rate, t_test_rate)
    if trisk_mod:
        fname = fname + '_trisk-{}'.format(trisk_mod)

    fname = fname + '_half-{}_vent-{}'\
        .format(bool_dict[half_classes], ventilation_mod)
    fname = fname + fname_addition + '.txt'

    with open(fname,'w')\
                   as outfile:
        json.dump(data, outfile)

def get_measures(measure_string, test_participation_rate=False,
        reduced_class_size=False, added_friendship_contacts=False,
        reduced_mask_efficiency=False, transmission_risk_modifier=False,
        vaccinations=False):
    '''
    Convenience function to get the individual measures given a string (filename)
    of measures
    '''
    #print(measure_string)
    agents = {
        'student':{
                'screening_interval': None, 
                'index_probability': 0, 
                'mask':False},
        'teacher':{
                'screening_interval': None, 
                'index_probability': 0, 
                'mask':False},
        'family_member':{
                'screening_interval': None, 
                'index_probability': 0, 
                'mask':False} 
}
    
    turnovers = {0:'same', 1:'one', 2:'two', 3:'three'}
    bmap = {'T':True, 'F':False}
    interval_map = {'0':0, '3':3, '7':7, '14':14, 'None':None}
    index_map = {'s':'student', 't':'teacher'}
    
    stype, _ = measure_string.split('_test')
    rest = measure_string.split(stype + '_')[1]

    if test_participation_rate and reduced_class_size and \
       reduced_mask_efficiency and transmission_risk_modifier:
        ttpype, turnover, index, tf, sf, tmask, smask, \
        class_size_reduction, vent, meffexh, meffinh, \
        s_test_rate, t_test_rate, trisk = rest.split('_')
        tmp = [stype, ttpype, turnover, index, tf, sf, tmask,
               smask, class_size_reduction, vent, meffexh, meffinh,
               s_test_rate, t_test_rate, trisk]
        
    elif test_participation_rate:
        ttpype, turnover, index, tf, sf, tmask, smask, haf, \
            s_test_rate, t_test_rate, vent = rest.split('_')
        tmp = [stype, ttpype, turnover, index, tf, sf, tmask,
         smask, haf, s_test_rate, t_test_rate, vent]
    elif reduced_class_size:
        ttpype, turnover, index, tf, sf, tmask, smask, haf, \
            class_size_reduction, vent = rest.split('_')
        tmp = [stype, ttpype, turnover, index, tf, sf, tmask,
         smask, haf, class_size_reduction, vent]
    elif added_friendship_contacts:
        ttpype, turnover, index, tf, sf, tmask, smask, haf, \
            friendship_contacts, vent = rest.split('_')
        tmp = [stype, ttpype, turnover, index, tf, sf,
            tmask, smask, haf, friendship_contacts, vent]
    elif reduced_mask_efficiency:
        ttpype, turnover, index, tf, sf, tmask, smask, \
        meffexh, meffinh, haf, vent = rest.split('_')
        tmp = [stype, ttpype, turnover, index, tf, sf,
            tmask, smask, haf, meffexh, meffinh, vent]
    else:
        if vaccinations:
            ttpype, turnover, index, tf, sf, tmask, smask, haf, vent, \
                tvacc, svacc = rest.split('_')
            tmp = [stype, ttpype, turnover, index, tf, sf, tmask, smask,
                   haf, vent, tvacc, svacc]
        else:
            ttpype, turnover, index, tf, sf, tmask, smask, haf, vent = \
                rest.split('_')
            tmp = [stype, ttpype, turnover, index, tf, sf, tmask, smask,
                   haf, vent]

    tmp = [m.split('-') for m in tmp]
    screening_params = {}
    
    half = False
    for m in tmp:
        if len(m) == 1:
            pass
        elif m[0] == 'test':
            ttype = '{}_day_{}'.format(turnovers[int(tmp[2][1])], tmp[1][1])
            screening_params['preventive_test_type'] = ttype
        elif m[0] == 'turnover':
            pass
        elif m[0] == 'index':
            screening_params['index_case'] = index_map[m[1]]
        elif m[0] == 'tf':
            agents['teacher']['screening_interval'] = interval_map[m[1]]
        elif m[0] == 'sf':
            agents['student']['screening_interval'] = interval_map[m[1]]
        elif m[0] == 'tmask':
            agents['teacher']['mask'] = bmap[m[1]]    
        elif m[0] == 'smask':
            agents['student']['mask'] = bmap[m[1]]
        elif m[0] == 'half':
            half = bmap[m[1]]
        elif m[0] == 'vent':
            screening_params['transmission_risk_ventilation_modifier'] = float(m[1])
        elif m[0] == 'csizered':
            screening_params['class_size_reduction'] = float(m[1])
        elif m[0] == 'stestrate':
            screening_params['student_test_rate'] = float(m[1])
        elif m[0] == 'ttestrate':
            screening_params['teacher_test_rate'] = float(m[1])
        elif m[0] == 'fcontacts':
            screening_params['added_friendship_contacts'] = float(m[1])
        elif m[0] == 'meffexh':
            screening_params['mask_efficiency_exhale'] = float(m[1])
        elif m[0] == 'meffinh':
            screening_params['mask_efficiency_inhale'] = float(m[1])
        elif m[0] == 'trisk':
            screening_params['transmission_risk_modifier'] = float(m[1])
        elif m[0] == 'tvacc':
            agents['teacher']['vaccination_ratio'] = float(m[1])
        elif m[0] == 'svacc':
            agents['student']['vaccination_ratio'] = float(m[1])
        else:
            print('unknown measure type ', m[0])
    
    return screening_params, agents, half


def get_data(stype, src_path, test_participation_rate=False,
            reduced_class_size=False, added_friendship_contacts=False,
            reduced_mask_efficiency=False, transmission_risk_modifier=False,
            vaccinations=False):
    '''
    Convenience function to read all ensembles from different measures
    of a given school type and return one single data frame
    '''
    data = pd.DataFrame()
    stype_path = join(src_path, stype)
    files = os.listdir(stype_path)
    for f in files:
        screening_params, agents, half = get_measures(f.strip('.csv'),
                test_participation_rate=test_participation_rate,
                reduced_class_size=reduced_class_size, 
                added_friendship_contacts=added_friendship_contacts,
                reduced_mask_efficiency=reduced_mask_efficiency,
                transmission_risk_modifier=transmission_risk_modifier,
                vaccinations=vaccinations)

        ensmbl = pd.read_csv(join(stype_path, f))
        try:
            ensmbl = ensmbl.drop(columns=['Unnamed: 0'])
        except KeyError:
            pass

        ensmbl['preventive_test_type'] = screening_params['preventive_test_type']
        ensmbl['index_case'] = screening_params['index_case']
        ensmbl['transmission_risk_ventilation_modifier'] = \
            screening_params['transmission_risk_ventilation_modifier']
        ensmbl['student_mask'] = agents['student']['mask']
        ensmbl['teacher_mask'] = agents['teacher']['mask']
        ensmbl['student_screening_interval'] = agents['student']['screening_interval']
        ensmbl['teacher_screening_interval'] = agents['teacher']['screening_interval']
        ensmbl['half_classes'] = half
        if reduced_class_size:
            ensmbl['class_size_reduction'] = screening_params['class_size_reduction']

        if test_participation_rate:
            ensmbl['student_test_rate'] = screening_params['student_test_rate']
            ensmbl['teacher_test_rate'] = screening_params['teacher_test_rate']

        if reduced_mask_efficiency:
            ensmbl['mask_efficiency_inhale'] = screening_params['mask_efficiency_inhale']
            ensmbl['mask_efficiency_exhale'] = screening_params['mask_efficiency_exhale']
            
        if transmission_risk_modifier:
            ensmbl['transmission_risk_modifier'] = screening_params['transmission_risk_modifier']
        
        if vaccinations:
            ensmbl['teacher_vaccination_ratio'] = agents['teacher']['vaccination_ratio']
            ensmbl['student_vaccination_ratio'] = agents['student']['vaccination_ratio']
        data = pd.concat([data, ensmbl])

    data = data.reset_index(drop=True)
    data['teacher_screening_interval'] = data['teacher_screening_interval'].replace({None:'never'})
    data['student_screening_interval'] = data['student_screening_interval'].replace({None:'never'})
    return data


import numpy as np
import networkx as nx
from math import gamma
from scipy.optimize import root_scalar

from mesa import Model
from mesa.time import RandomActivation, SimultaneousActivation
from mesa.datacollection import DataCollector

from scseirx.testing_strategy import Testing

## data collection functions ##
def get_N_diagnostic_tests(model):
    return model.number_of_diagnostic_tests


def get_N_preventive_screening_tests(model):
    return model.number_of_preventive_screening_tests


def get_infection_state(agent):
    if agent.exposed == True: return 'exposed'
    elif agent.infectious == True: return 'infectious'
    elif agent.recovered == True: return 'recovered'
    else: return 'susceptible'

def get_quarantine_state(agent):
    if agent.quarantined == True: return True
    else: return False


def get_undetected_infections(model):
    return model.undetected_infections


def get_predetected_infections(model):
    return model.predetected_infections


def get_pending_test_infections(model):
    return model.pending_test_infections


def get_diagnostic_test_detected_infections_student(model):
    return model.positive_tests[model.Testing.diagnostic_test_type]['student']


def get_diagnostic_test_detected_infections_teacher(model):
    return model.positive_tests[model.Testing.diagnostic_test_type]['teacher']


def get_diagnostic_test_detected_infections_family_member(model):
    return model.positive_tests[model.Testing.diagnostic_test_type]\
        ['family_member']


def get_preventive_test_detected_infections_student(model):
    return model.positive_tests[model.Testing.preventive_screening_test_type]\
        ['student']


def get_preventive_test_detected_infections_teacher(model):
    return model.positive_tests[model.Testing.preventive_screening_test_type]\
        ['teacher']


def get_preventive_test_detected_infections_family_member(model):
    return model.positive_tests[model.Testing.preventive_screening_test_type]\
        ['family_member']


# parameter sanity check functions


def check_positive(var):
	assert var >= 0, 'negative number'
	return var


def check_bool(var):
	assert type(var) == bool, 'not a bool'
	return var


def check_positive_int(var):
    if var == None:
        return var
    assert type(var) == int, 'not an integer'
    assert var >= 0, 'negative number'
    return var


def check_contact_type_dict(var):
	assert type(var) == dict, 'not a dictionary'
	assert set(var.keys()).issubset({'very_far', 'far', 'intermediate', 'close'}), \
		'does not contain the correct contact types (has to be very_far, far, intermediate or close)'
	assert all((isinstance(i, int) or isinstance(i, float)) for i in var.values()), \
		'contact type weights are not numeric'

	return var


def check_K1_contact_types(var):
    for area in var:
        assert area in ['very_far', 'far', 'intermediate',
            'close'], 'K1 contact type not recognised'
    return var


def check_testing(var):
    assert var in ['diagnostic', 'background', 'preventive',
                   'background+preventive', False], \
        'unknown testing mode: {}'.format(var)

    return var



def check_probability(var):
	assert (type(var) == float) or (var == 0) or (var == 1), \
		 '{} not a float'.format(var)
	assert var >= 0, 'probability negative'
	assert var <= 1, 'probability larger than 1'
	return var


def check_graph(var):
    assert type(var) in [nx.Graph, nx.MultiGraph], 'not a networkx graph'
    assert len(var.nodes) > 0, 'graph has no nodes'
    assert len(var.edges) > 0, 'graph has no edges'
    areas = [e[2]['contact_type'] for e in var.edges(data=True)]
    areas = set(areas)
    for a in areas:
        assert a in {'very_far', 'far', 'intermediate',
            'close'}, 'contact type {} not recognised'.format(a)
    return var


def check_index_case(var, agent_types):
	allowed_strings = agent_types[:]
	allowed_strings.extend(['continuous'])
	assert var in allowed_strings, 'unknown index case mode'
	return var


def check_discount(var):
    if var['slope'] != None:
        assert var['slope'] <= 0, 'slope needs to be <= 0 or None'
        assert np.abs(var['slope']) <= 1, 'absolute value of slope needs to be <= 1'
    assert var['intercept'], 'intercept needs to be positive'
    assert var['intercept'], 'intercept needs to be <= 1'
    return var


def get_weibull_shape(k, mu, var):
    '''
    Calculates the shape parameter of a Weibull distribution, given its mean
    mu and its variance var
    '''
    return var / mu**2 - gamma(1 + 2/k) / gamma(1+1/k)**2 + 1



def get_weibull_scale(mu, k):
    '''
    Calculates the scale parameter of a Weibull distribution, given its mean
    mu and its shape parameter k
    '''
    return mu / gamma(1 + 1/k)


def weibull_two_param(shape, scale):
    '''
    A two-parameter Weibull distribution, based on numpy ramdon's single
    parameter distribution. We use this distribution in the simulation to draw
    random epidemiological parameters for agents from the given distribution
    See https://numpy.org/doc/stable/reference/random/generated/numpy.random.weibull.html
    '''
    return scale * np.random.weibull(shape)


class SEIRX(Model):
    '''
    A model with a number of different agents that reproduces
    the SEIRX dynamics of pandemic spread in a facility. Note:
    all times are set to correspond to days

    G: networkx undirected graph, interaction graph between agents. Edges have
    to have edge the edge attribute 'contact_type' specifying the closeness of
    contacts, which can be ['very far', 'far', 'intermediate' and 'close'].
    Nodes have to have the node attribute 'type' which specifies the agent type
    of the given node (for example 'student' or 'teacher' in a school scenario).
    In addition, nodes can have the attribute 'unit', which assigns them to a
    unit in space (for example a 'class' in a school scenario).

    verbosity: integer in [0, 1, 2], controls text output to std out to track
    simulation progress and transmission dynamics. Default = 0.

    testing, default = 'diagnostic'
        'diagnostic':   only diagnostic tests for symptomatic agents
        'background':   adds background screens of all agents after a positive
                        diagnostic test
        'preventive':   adds preventive screens of agent groups to diagnostic
                        testing. Screens happen in time intervals specified 
                        separately for each agent group in the variable 
                        'screening_interval'.
        'background+preventive': preventive screens AND background screens on
                        top of diagnostic testing.

    infection_duration, default = 11 NOTE: includes the time an agent is exposed
    but not yet infectious at the beginning of an infection
        positive integer:   mean or median of the infection duration in days
        list of two floats: mean and standard deviation of a distribution
                            specifying the infection duration in days. These
                            numbers will be used to construct a Weibull
                            distribution from which the infection duration will
                            be drawn for every agent individually

    exposure_duration, default = 4. Sets the time from transmission to becoming
    infectious
        positive integer:   mean or median of the exposure duration in days
        list of two floats: mean and standard deviation of a distribution
                            specifying the exposure duration in days. These
                            numbers will be used to construct a Weibull
                            distributoin from which the exposure duration will
                            be drawn for every agent individually.

    time_until_symptoms, default = 6. Sets the time from transmission to
    (potentially) developing symptoms. Symptom probability has to be set for
    each agent group individually using the parameter 'symptom_probability'
        positive integer:   mean or median of the time until symptoms in days
        list of two floats: mean and standard deviation of a distribution
                            specifying the time until symptoms in days. These
                            numbers will be used to construct a Weibull
                            distribution from which the time until symptoms will
                            be drawn for every agent individually.

    quarantine_duration, default = 14. Positive integer, sets the time a
    positively tested agent is quarantined in days

    infection_risk_contact_type_weights: dictionary of the form
    {'very_far':float, 'far':float, 'intermediate':float, 'close':float}
    that sets transmission risk multipliers for different contact types of
    agents specified in the contact network G. Default: {'very_far': 0.1,
    'far': 0.5, 'intermediate': 1, 'close': 3}

    subclinical_modifier: default = 1.0. Float, modifies the infectiousness of
    asymptomatic cases. Example: if subclinical_modifier = 0.5, the
    infectiousness of an asymptomatic case will be reduced to 50%.

    K1_contact_types: list of strings from ['very_far', 'far', 'intermediate',
    'close']. Definition of contact types for which agents are considered
    "K1 contact persons" if they had contact to a positively tested person wtith
    a specified contact intensity. Default = ['close'].

    diagnostic_test_type, default = 'one_day_PCR'. String, specifies the test
    technology and test result turnover time used for diagnostic testing. For
    example 'same_day_antigen' or 'two_day_PCR'. See module "Testing" for
    different implemented testing techologies.

    preventive_screening_test_type:, default = 'one_day_PCR', String, specifies
    the test technology and test result turnover time used for preventive
    sreening. For example 'same_day_antigen' or 'two_day_PCR'. See module
    "Testing" for different implemented testing techologies.

    follow_up_testing_interval, default = None. Positive integer, sets the time
    a follow-up screen (background screen) is initiated after an initial screen
    triggered by a positive test result. Only applies if the testing strategy is
    'background' or preventive.

    liberating_testing, default = False. Boolean, flag that specifies, whether
    or not an agent is released from quarantine after returning a negative test
    result.

	index_case, default = 'employee' (nursing home scenario) or 'teacher'
    (school scenario). Specifies how infections are introduced into the facility.
        agent_type:     If an agent type (for example 'student' or 'teacher' in
                        the school scenario) is specified, a single randomly
                        chosen agent from this agent group will become the index
                        case and no further index cases will be introduced into
                        the scenario.
        'continuous':   In this case, agents have a continuous risk to become
                        index cases in every simulation step. The risk has to
                        be specified for every agent group individually, using
                        the 'index_probability' parameter. If only a single
                        agent group has a non-zero index probability, then only
                        agents from this group can become index cases.


    agent_types: dictionary of the structure
        {
        agent type:
            {
            screening interval : integer, number of days between each preventive
            screen in this agent group

            index probability : float in the range [0, 1], sets the probability
            to become an index case in each time step

            mask : bool
                whether or not the agent type is wearing a mask
            }
        }

    The dictionary's keys are the names of the agent types which have to
    correspond to the node attributes in the contact graph. The screening
    interval sets the time-delay between preventive screens of this agent group,
    the index probability sets the probability of a member of this agent group
    becoming an index case in every time step

    seed: positive integer, fixes the seed of the simulation to enable
    repeatable simulation runs. If seed = None, the simulation will be
    initialized at random.
    '''

    def __init__(self, G,
        verbosity = 0,
        base_transmission_risk = 0.05,
        testing='diagnostic',
        exposure_duration = [5.0, 1.9],
        time_until_symptoms = [6.4, 0.8],
        infection_duration = [10.91, 3.95],
        #vaccinated = False,
        quarantine_duration = 10,
        subclinical_modifier = 0.6,
        infection_risk_contact_type_weights = {
            'very_far': 0.1,
            'far': 0.25,
            'intermediate': 0.5,
            'close': 1},
        K1_contact_types = ['close'],
        diagnostic_test_type = 'one_day_PCR',
        preventive_screening_test_type = 'same_day_antigen',
        follow_up_testing_interval = None,
        liberating_testing = False,
        index_case = 'teacher',
        agent_types = {
            'teacher':      {'screening_interval': None,
                             'index_probability': 0,
                             'mask':False,
                             'vaccination_ratio': 0},
            'student':      {'screening_interval': None,
                             'index_probability': 0,
                             'mask':False,
                             'vaccination_ratio': 0},
            'family_member':{'screening_interval': None,
                             'index_probability': 0,
                             'mask':False,
                             'vaccination_ratio': 0}},
        age_transmission_risk_discount = \
             {'slope':-0.02,
              'intercept':1},
        age_symptom_modification = \
             {'slope':-0.02545,
              'intercept':0.854545},
        mask_filter_efficiency = {'exhale':0, 'inhale':0},
        transmission_risk_ventilation_modifier = 0,
        transmission_risk_vaccination_modifier = {
            'reception':1,
            'transmission':0},
        seed = None):

        # mesa models already implement fixed seeds through their own random
        # number generations. Sadly, we need to use the Weibull distribution
        # here, which is not implemented in mesa's random number generation
        # module. Therefore, we need to initialize the numpy random number
        # generator with the given seed as well
        if seed != None:
            np.random.seed(seed)

        # sets the (daily) transmission risk for a household contact without
        # any precautions. Target infection ratios are taken from literature
        # and the value of the base_transmission_risk is calibrated such that
        # the simulation produces the correct infection ratios in a household
        # setting with the given distributions for epidemiological parameters
        # of agents
        self.base_transmission_risk = base_transmission_risk
    	# sets the level of detail of text output to stdout (0 = no output)
        self.verbosity = check_positive_int(verbosity)
        # flag to turn off the testing & tracing strategy
        self.testing = check_testing(testing)
        self.running = True  # needed for the batch runner implemented by mesa
        # set the interaction mode to simultaneous activation
        self.schedule = SimultaneousActivation(self)


        # internal step counter used to launch screening tests
        self.Nstep = 0

        # since we may have weekday-specific contact networks, we need
        # to keep track of the day of the week. Since the index case
        # per default is introduced at step 0 in index case mode, we
        # need to offset the starting weekday by a random number of weekdays
        # to prevent artifacts from always starting on the same day of the week

        self.weekday_offset = self.random.randint(1, 8)
        self.weekday = self.Nstep + self.weekday_offset

        ## epidemiological parameters: can be either a single integer or the
        # mean and standard deviation of a distribution
        self.epi_params = {}
        # counter to track the number of pathological parameter combinations
        # that had to be re-rolled (only here for debugging and control reasons)
        self.param_rerolls = 0

        for param, param_name in zip([exposure_duration, time_until_symptoms,
                infection_duration],['exposure_duration', 'time_until_symptoms',
                'infection_duration']):

            if isinstance(param, int):
                self.epi_params[param_name] = check_positive_int(param)

            elif isinstance(param, list) and len(param) == 2:

                mu = check_positive(param[0])
                var = check_positive(param[1]**2)
                shape = root_scalar(get_weibull_shape, args=(mu, var),
                            method='toms748', bracket=[0.2, 500]).root
                scale = get_weibull_scale(mu, shape)

                self.epi_params[param_name] = [shape, scale]
            else:
                print('{} format not recognized, should be either a single '+\
                  'int or a tuple of two positive numbers'.format(param_name))


        # duration of quarantine
        self.quarantine_duration = check_positive_int(quarantine_duration)

        self.infection_risk_area_weights = check_contact_type_dict(
            infection_risk_contact_type_weights)

        # modifier for infectiosness for asymptomatic cases
        self.subclinical_modifier = check_positive(subclinical_modifier)
        # modifiers for the infection risk, depending on contact type
        self.infection_risk_contact_type_weights = infection_risk_contact_type_weights

        # modifications for age-dependent transmission and reception risks and
        # symptom probabilities
        self.age_transmission_risk_discount = \
            check_discount(age_transmission_risk_discount)

        self.age_symptom_modification = age_symptom_modification
            #check_discount(age_symptom_modification)

        self.mask_filter_efficiency = mask_filter_efficiency
        self.transmission_risk_ventilation_modifier = \
            transmission_risk_ventilation_modifier
        self.transmission_risk_vaccination_modifier = \
            transmission_risk_vaccination_modifier
        ## agents and their interactions
        # interaction graph of agents
        self.G = check_graph(G)
        # add weights as edge attributes so they can be visualised easily
        if type(self.G) == nx.MultiGraph:
            for (u, v, key, contact_type) in self.G.edges(keys=True,
                    data='contact_type'):
                self.G[u][v][key]['weight'] = \
                    self.infection_risk_contact_type_weights[contact_type]
        else:
            for e in G.edges(data=True):
                G[e[0]][e[1]]['weight'] = self.infection_risk_contact_type_weights\
                	[G[e[0]][e[1]]['contact_type']]

        # extract the different agent types from the contact graph
        self.agent_types = list(agent_types.keys())
        # dictionary of available agent classes with agent types and classes
        self.agent_classes = {}
        if 'resident' in agent_types:
            from scseirx.agent_resident import resident
            self.agent_classes['resident'] = resident
        if 'employee' in agent_types:
            from scseirx.agent_employee import employee
            self.agent_classes['employee'] = employee
        if 'student' in agent_types:
            from scseirx.agent_student import student
            self.agent_classes['student'] = student
        if 'teacher' in agent_types:
            from scseirx.agent_teacher import teacher
            self.agent_classes['teacher'] = teacher
        if 'family_member' in agent_types:
            from scseirx.agent_family_member import family_member
            self.agent_classes['family_member'] = family_member

        ## set agent characteristics for all agent groups
        # list of agent characteristics
        params = ['screening_interval','index_probability', 'mask' ,'vaccination_ratio',
                  'voluntary_testing_rate']

        # default values that are used in case a characteristic is not specified
        # for an agent group
        defaults = {'screening_interval':None,
                    'index_probability':0,
                    'mask':False,
                    'vaccination_ratio':0,
                    'voluntary_testing_rate':1
                    }

        # sanity checks that are applied to parameters passed to the class
        # constructor to make sure they conform to model expectations
        check_funcs = [check_positive_int, check_probability, check_bool,
                       check_probability, check_probability]

        # member dicts that store the parameter values for each agent group
        self.screening_intervals = {}
        self.index_probabilities = {}
        self.masks = {}
        self.vaccination_probabilities = {}
        self.voluntary_testing_rates = {}


        param_dicts = [self.screening_intervals, self.index_probabilities,
                    self.masks, self.vaccination_probabilities, self.voluntary_testing_rates]

        # iterate over all possible agent parameters and agent groups: set the
        # respective value to the value passed through the constructor or to
        # the default value if no value has been passed
        for param,param_dict,check_func in zip(params,param_dicts,check_funcs):
            for at in self.agent_types:
                try:
                    param_dict.update({at:check_func(agent_types[at][param])})
                except KeyError:
                    param_dict.update({at:defaults[param]})

        # pass all parameters relevant for the testing strategy to the testing
        # class. NOTE: this separation is not a strictly necessary design
        # decision but I like to keep the parameters related to testing and
        # tracing in a separate place
        self.Testing = Testing(self, diagnostic_test_type,
             preventive_screening_test_type,
             check_positive_int(follow_up_testing_interval),
             self.screening_intervals,
             check_bool(liberating_testing),
             check_K1_contact_types(K1_contact_types),
             verbosity)


        # specifies either continuous probability for index cases in agent
        # groups based on the 'index_probability' for each agent group, or a
        # single (randomly chosen) index case in the passed agent group
        self.index_case = check_index_case(index_case, self.agent_types)

        self.num_agents = {}

        ## add agents
        # extract the agent nodes from the graph and add them to the scheduler
        for agent_type in self.agent_types:
            IDs = [x for x,y in G.nodes(data=True) if y['type'] == agent_type]
            self.num_agents.update({agent_type:len(IDs)})

            # get the agent locations (units) from the graph node attributes
            units = [self.G.nodes[ID]['unit'] for ID in IDs]

            # determine the agents that will be vaccinated, given the 
            # vaccination ratio of the respective agent group
            vaccination_status = np.asarray([False] * len(IDs))
            if self.vaccination_probabilities[agent_type] > 0:
                n = round(self.vaccination_probabilities[agent_type] * len(IDs))
                idx = list(range(len(IDs)))
                rnd_idx = np.asarray(self.random.sample(idx, n))
                vaccination_status[rnd_idx] = True


            for ID, unit, vaccinated in zip(IDs, units, vaccination_status):

                tmp_epi_params = {}
                # for each of the three epidemiological parameters, check if
                # the parameter is an integer (if yes, pass it directly to the
                # agent constructor), or if it is specified by the shape and
                # scale parameters of a Weibull distribution. In the latter
                # case, draw a new number for every agent from the distribution
                # NOTE: parameters drawn from the distribution are rounded to
                # the nearest integer
                while True:
                    for param_name, param in self.epi_params.items():
                        if isinstance(param, int):
                            tmp_epi_params[param_name] = param

                        else:
                            tmp_epi_params[param_name] = \
                                round(weibull_two_param(param[0], param[1]))

                    if tmp_epi_params['exposure_duration'] > 0 and \
                       tmp_epi_params['time_until_symptoms'] >= \
                       tmp_epi_params['exposure_duration'] and\
                       tmp_epi_params['infection_duration'] > \
                       tmp_epi_params['exposure_duration']:
                       break
                    else:
                        self.param_rerolls += 1
                        if verbosity > 1:
                            print('pathological epi-param case found!')
                            print(tmp_epi_params)

                # check if the agent participates in voluntary testing
                p = self.voluntary_testing_rates[agent_type]
                voluntary_testing = np.random.choice([True, False],
                         p=[p, 1-p])

                # construct the agent object
                a = self.agent_classes[agent_type](ID, unit, self,
                    tmp_epi_params['exposure_duration'],
                    tmp_epi_params['time_until_symptoms'],
                    tmp_epi_params['infection_duration'],
                    vaccinated,
                    voluntary_testing,
                    verbosity)
                self.schedule.add(a)


		# infect the first agent in single index case mode
        if self.index_case != 'continuous':
            infection_targets = [
                a for a in self.schedule.agents if a.type == index_case]
            # pick a random agent to infect in the selected agent group
            target = self.random.randint(0, len(infection_targets) - 1)
            infection_targets[target].exposed = True
            if self.verbosity > 0:
                print('{} exposed: {}'.format(index_case,
                    infection_targets[target].ID))


        # list of agents that were tested positive this turn
        self.newly_positive_agents = []
        # flag that indicates if there were new positive tests this turn
        self.new_positive_tests = False
        # dictionary of flags that indicate whether a given agent group has
        # been creened this turn
        self.screened_agents= {
            'reactive':{agent_type: False for agent_type in self.agent_types},
            'follow_up':{agent_type: False for agent_type in self.agent_types},
            'preventive':{agent_type: False for agent_type in self.agent_types}}


        # dictionary of counters that count the days since a given agent group
        # was screened. Initialized differently for different index case modes
        if (self.index_case == 'continuous') or \
      	   (not np.any(list(self.Testing.screening_intervals.values()))):
        	self.days_since_last_agent_screen = {agent_type: 0 for agent_type in
        	self.agent_types}
        # NOTE: if we initialize these variables with 0 in the case of a single
        # index case, we introduce a bias since in 'single index case mode' the
        # first index case will always become exposed in step 0. To realize
        # random states of the preventive sceening procedure with respect to the
        # incidence of the index case, we have to randomly pick the days since
        # the last screen for the agent group from which the index case is
        else:
        	self.days_since_last_agent_screen = {}
        	for agent_type in self.agent_types:
        		if self.Testing.screening_intervals[agent_type] != None:
        			self.days_since_last_agent_screen.update({
        				agent_type: self.random.choice(range(0,
        				 self.Testing.screening_intervals[agent_type] + 1))})
        		else:
        			self.days_since_last_agent_screen.update({agent_type: 0})

        # dictionary of flags that indicates whether a follow-up screen for a
        # given agent group is scheduled
        self.scheduled_follow_up_screen = {agent_type: False for agent_type in
        	self.agent_types}

        # counters
        self.number_of_diagnostic_tests = 0
        self.number_of_preventive_screening_tests = 0
        self.positive_tests = {self.Testing.preventive_screening_test_type:
                            {agent_type:0 for agent_type in self.agent_types},
                               self.Testing.diagnostic_test_type:
                            {agent_type:0 for agent_type in self.agent_types}}

        self.undetected_infections = 0
        self.predetected_infections = 0
        self.pending_test_infections = 0
        self.quarantine_counters = {agent_type:0 for agent_type in agent_types.keys()}
        self.false_negative = 0

        # data collectors to save population counts and agent states every
        # time step
        self.datacollector = DataCollector(
            model_reporters=
            	{
            	'N_diagnostic_tests':get_N_diagnostic_tests,
                'N_preventive_screening_tests':get_N_preventive_screening_tests,
                'diagnostic_test_detected_infections_student':\
                        get_diagnostic_test_detected_infections_student,
                'diagnostic_test_detected_infections_teacher':\
                        get_diagnostic_test_detected_infections_teacher,
                'diagnostic_test_detected_infections_family_member':\
                        get_diagnostic_test_detected_infections_family_member,
                'preventive_test_detected_infections_student':\
                        get_preventive_test_detected_infections_student,
                'preventive_test_detected_infections_teacher':\
                        get_preventive_test_detected_infections_teacher,
                'preventive_test_detected_infections_family_member':\
                        get_preventive_test_detected_infections_family_member,
                'undetected_infections':get_undetected_infections,
                'predetected_infections':get_predetected_infections,
                'pending_test_infections':get_pending_test_infections
                },

            agent_reporters=
            	{
            	'infection_state': get_infection_state,
                'quarantine_state': get_quarantine_state
                })


    ## transmission risk modifiers
    def get_transmission_risk_contact_type_modifier(self, source, target):
        # construct the edge key as combination between agent IDs and weekday
        n1 = source.ID
        n2 = target.ID
        tmp = [n1, n2]
        tmp.sort()
        n1, n2 = tmp
        key = n1 + n2 + 'd{}'.format(self.weekday)
        contact_weight = self.G.get_edge_data(n1, n2, key)['weight']

        # the link weight is a multiplicative modifier of the link strength.
        # contacts of type "close" have, by definition, a weight of 1. Contacts
        # of type intermediate, far or very far have a weight < 1 and therefore
        # are less likely to transmit an infection. For example, if the contact
        # type far has a weight of 0.2, a contact of type far has only a 20%
        # chance of transmitting an infection, when compared to a contact of
        # type close. To calculate the probability of success p in the Bernoulli
        # trial, we need to reduce the base risk (or base probability of success)
        # by the modifications introduced by preventive measures. These
        # modifications are formulated in terms of "probability of failure", or
        # "q". A low contact weight has a high probability of failure, therefore
        # we return q = 1 - contact_weight here.
        q1 = 1 - contact_weight

        return q1


    def get_transmission_risk_age_modifier_transmission(self, source):
        '''linear function such that at age 18 the risk is that of an adult (=1).
        The slope of the line needs to be calibrated.
        '''
        age = source.age
        max_age = 18
        if age <= max_age:
            age_weight = self.age_transmission_risk_discount['slope'] * \
                 np.abs(age - max_age) + self.age_transmission_risk_discount['intercept']

            # The age weight can be interpreted as multiplicative factor that
            # reduces the chance for transmission with decreasing age. The slope
            # of the age_transmission_discount function is the decrease (in % of
            # the transmission risk for an 18 year old or above) of transmission
            # risk with every year a person is younger than 18 (the intercept is
            # 1 by definition).
            # To calculate the probability of success p in the Bernoulli
            # trial, we need to reduce the base risk (or base probability of 
            # success) by the modifications introduced by preventive measures. 
            # These modifications are formulated in terms of "probability of 
            # failure", or "q". A low age weight has a high probability of 
            # failure, therefore we return q = 1 - age_weight here.
            q2 = 1 - age_weight
        else:
            q2 = 0

        return q2


    def get_transmission_risk_age_modifier_reception(self, target):
        '''linear function such that at age 18 the risk is that of an adult (=1).
        The slope of the line needs to be calibrated.
        '''
        age = target.age
        max_age = 18
        if age <= max_age:
            age_weight = self.age_transmission_risk_discount['slope'] * \
            np.abs(age - max_age) + self.age_transmission_risk_discount['intercept']
            # see description in get_transmission_risk_age_modifier_transmission
            q3 = 1 - age_weight
        else:
            q3 = 0

        return q3


    # infectiousness is constant and high until symptom onset and then
    # decreases monotonically until agents are not infectious anymore
    # at the end of the infection_duration
    def get_transmission_risk_progression_modifier(self, source):
        if source.days_since_exposure < source.exposure_duration:
            progression_weight = 0
        elif source.days_since_exposure <= source.time_until_symptoms:
            progression_weight = 1
        elif source.days_since_exposure > source.time_until_symptoms and \
             source.days_since_exposure <= source.infection_duration:
            # we add 1 in the denominator, such that the source is also
            # (slightly) infectious on the last day of the infection_duration
            progression_weight = \
                 (source.days_since_exposure - source.time_until_symptoms) / \
                 (source.infection_duration - source.time_until_symptoms + 1)
        else:
            progression_weight = 0
        # see description in get_transmission_risk_age_modifier_transmission
        q4 = 1 - progression_weight

        return q4

    def get_transmission_risk_subclinical_modifier(self, source):
        if source.symptomatic_course == False:
            subclinical_weight = self.subclinical_modifier
        else:
            subclinical_weight = 1
        # see description in get_transmission_risk_age_modifier_transmission
        q5 = 1 - subclinical_weight
        return q5

    def get_transmission_risk_exhale_modifier(self, source):
        if source.mask:
            exhale_weight = self.mask_filter_efficiency['exhale']
        else:
            exhale_weight = 1
        # see description in get_transmission_risk_age_modifier_transmission
        q6 = 1 - exhale_weight
        return q6


    def get_transmission_risk_inhale_modifier(self, target):
        if target.mask:
            inhale_weight = self.mask_filter_efficiency['inhale']
        else:
            inhale_weight = 1
        # see description in get_transmission_risk_age_modifier_transmission
        q7 = 1 - inhale_weight
        return q7


    def get_transmission_risk_ventilation_modifier(self):
        ventilation_weight = self.transmission_risk_ventilation_modifier
        # see description in get_transmission_risk_age_modifier_transmission
        q8 = 1 - ventilation_weight
        return q8

    def get_transmission_risk_vaccination_modifier_reception(self, a):
        if a.vaccinated:
            q9 = self.transmission_risk_vaccination_modifier['reception']
        else:
            q9 = 0
        return q9

    def get_transmission_risk_vaccination_modifier_transmission(self, a):
        if a.vaccinated:
            q10 = self.transmission_risk_vaccination_modifier['transmission']
        else:
            q10 = 0
        return q10

    def test_agent(self, a, test_type):
        a.tested = True
        a.pending_test = test_type
        if test_type == self.Testing.diagnostic_test_type:
            self.number_of_diagnostic_tests += 1
        else:
            self.number_of_preventive_screening_tests += 1

        if a.exposed:
            # tests that happen in the period of time in which the agent is
            # exposed but not yet infectious. 
            # Note: tests[test_type]['time_until_testable'] is negative for
            # tests that can detect an infection before agents become infectious
            if a.days_since_exposure >= a.exposure_duration + \
                    self.Testing.tests[test_type]['time_until_testable']:
                
                if self.verbosity > 1:
                    print('{} {} sent positive sample (even though not infectious yet)'
                    .format(a.type, a.ID))
                a.sample = 'positive'
                self.predetected_infections += 1
                self.positive_tests[test_type][a.type] += 1
            else:
                if self.verbosity > 1: print('{} {} sent negative sample'
                    .format(a.type, a.ID))
                a.sample = 'negative'

        elif a.infectious:
            # tests that happen in the period of time in which the agent is
            # infectious and the infection is detectable by a given test
            # Note: tests[test_type]['time_until_testable'] is negative for 
            # tests that can detect an infection before agents become 
            # infectious. tests[test_type]['time_testable'] is negative for
            # tests that cease to detect an infection before agents stop being
            # infectious
            if a.days_since_exposure >= a.exposure_duration + \
                    self.Testing.tests[test_type]['time_until_testable'] and \
               a.days_since_exposure <= a.infection_duration + \
                    self.Testing.tests[test_type]['time_testable']:
                if self.verbosity > 1:
                    print('{} {} sent positive sample'.format(a.type, a.ID))
                a.sample = 'positive'
                self.positive_tests[test_type][a.type] += 1

            # track the undetected infections to assess how important they are
            # for infection spread
            else:
                if self.verbosity > 1:
                    print('{} {} sent negative sample (even though infectious)'
                    .format(a.type, a.ID))
                a.sample = 'negative'
                self.undetected_infections += 1

        else:
            if self.verbosity > 1: print('{} {} sent negative sample'
                .format(a.type, a.ID))
            a.sample = 'negative'

        # for same-day testing, immediately act on the results of the test
        if a.days_since_tested >= self.Testing.tests[test_type]['time_until_test_result']:
            a.act_on_test_result()

    def screen_agents(self, agent_group, test_type, screen_type):
        # only test agents that have not been tested already in this simulation
        # step and that are not already known positive cases

        if self.verbosity > 0:
            print('initiating {} {} screen'\
                                .format(screen_type, agent_group))

        untested_agents = [a for a in self.schedule.agents if
            (a.tested == False and a.known_positive == False
                and a.type == agent_group)]

        if len(untested_agents) > 0:
            self.screened_agents[screen_type][agent_group] = True
            self.days_since_last_agent_screen[agent_group] = 0

            # only test agents if they participate in voluntary testing
            if screen_type == 'preventive':
                for a in untested_agents:
                    if a.voluntary_testing:
                        self.test_agent(a, test_type)
                    else:
                        if self.verbosity > 1:
                            print('not testing {} {}, not participating in voluntary testing'\
                                .format(agent_group, a.ID))
            else:
                for a in untested_agents:
                    self.test_agent(a, test_type)

            if self.verbosity > 0:
                print()
        else:
            if self.verbosity > 0:
                print('no agents tested because all agents have already been tested')

    # the type of the test used in the pending test result is stored in the
    # variable pending_test

    def collect_test_results(self):
        agents_with_test_results = [a for a in self.schedule.agents if
            (a.pending_test and
             a.days_since_tested >= self.Testing.tests[a.pending_test]['time_until_test_result'])]

        return agents_with_test_results

    def trace_contacts(self, a):
        if a.quarantined == False:
            a.quarantined = True
            a.quarantine_start = self.Nstep

            if self.verbosity > 0:
                print('qurantined {} {}'.format(a.type, a.ID))

        # find all agents that share edges with the agent
        # that are classified as K1 contact types in the testing
        # strategy
        K1_contacts = [e[1] for e in self.G.edges(a.ID, data=True) if
            e[2]['contact_type'] in self.Testing.K1_contact_types]
        K1_contacts = [a for a in self.schedule.agents if a.ID in K1_contacts]

        for K1_contact in K1_contacts:
            if self.verbosity > 0:
                print('quarantined {} {} (K1 contact of {} {})'
                    .format(K1_contact.type, K1_contact.ID, a.type, a.ID))
            K1_contact.quarantined = True
            K1_contact.quarantine_start = self.Nstep

    def test_symptomatic_agents(self):
        # find symptomatic agents that have not been tested yet and are not
        # in quarantine and test them
        newly_symptomatic_agents = np.asarray([a for a in self.schedule.agents
            if (a.symptoms == True and a.tested == False and a.quarantined == False)])

        for a in newly_symptomatic_agents:
            # all symptomatic agents are quarantined by default
            if self.verbosity > 0:
                print('quarantined: {} {}'.format(a.type, a.ID))
            a.quarantined = True
            a.quarantine_start = self.Nstep

            self.test_agent(a, self.Testing.diagnostic_test_type)

    def quarantine_contacts(self):
        # trace and quarantine contacts of newly positive agents
        if len(self.newly_positive_agents) > 0:
            if self.verbosity > 0: print('new positive test(s) from {}'
                .format([a.ID for a in self.newly_positive_agents]))

            # send all K1 contacts of positive agents into quarantine
            for a in self.newly_positive_agents:
                self.trace_contacts(a)

            # indicate that a screen should happen because there are new
            # positive test results
            self.new_positive_tests = True
            self.newly_positive_agents = []

        else:
            self.new_positive_tests = False


    def step(self):
        self.weekday = (self.Nstep + self.weekday_offset) % 7 + 1
        # if the connection graph is time-resloved, set the graph that is
        # used to determine connections in this step to the sub-graph corres-
        # ponding to the current day of the week
        if self.dynamic_connections:
            self.G = self.weekday_connections[self.weekday]

        if self.verbosity > 0:
            print('weekday {}'.format(self.weekday))

        if self.testing:
            for agent_type in self.agent_types:
                for screen_type in ['reactive', 'follow_up', 'preventive']:
                    self.screened_agents[screen_type][agent_type] = False

            if self.verbosity > 0:
                print('* testing and tracing *')

            self.test_symptomatic_agents()


            # collect and act on new test results
            agents_with_test_results = self.collect_test_results()
            for a in agents_with_test_results:
                a.act_on_test_result()

            self.quarantine_contacts()

            # screening:
            # a screen should take place if
            # (a) there are new positive test results
            # (b) as a follow-up screen for a screen that was initiated because
            # of new positive cases
            # (c) if there is a preventive screening policy and it is time for
            # a preventive screen in a given agent group

            # (a)
            if (self.testing == 'background' or self.testing == 'background+preventive')\
               and self.new_positive_tests == True:
                for agent_type in self.screening_agents:
                    self.screen_agents(
                        agent_type, self.Testing.diagnostic_test_type, 'reactive')
                    self.scheduled_follow_up_screen[agent_type] = True

            # (b)
            elif (self.testing == 'background' or self.testing == 'background+preventive') and \
                self.Testing.follow_up_testing_interval != None and \
                sum(list(self.scheduled_follow_up_screen.values())) > 0:
                for agent_type in self.screening_agents:
                    if self.scheduled_follow_up_screen[agent_type] and\
                       self.days_since_last_agent_screen[agent_type] >=\
                       self.Testing.follow_up_testing_interval:
                        self.screen_agents(
                            agent_type, self.Testing.diagnostic_test_type, 'follow_up')
                    else:
                        if self.verbosity > 0:
                            print('not initiating {} follow-up screen (last screen too close)'\
                                .format(agent_type))

            # (c) 
            elif (self.testing == 'preventive' or self.testing == 'background+preventive')and \
                np.any(list(self.Testing.screening_intervals.values())):

                for agent_type in self.screening_agents:
                    interval = self.Testing.screening_intervals[agent_type]
                    assert interval in [7, 3, 2, None], \
                        'testing interval {} for agent type {} not supported!'\
                        .format(interval, agent_type)

                    # (c.1) testing every 7 days = testing on Mondays
                    if interval == 7 and self.weekday == 1:
                        self.screen_agents(agent_type,
                            self.Testing.preventive_screening_test_type,\
                             'preventive')
                    # (c.2) testing every 3 days = testing on Mo & Turs
                    elif interval == 3 and self.weekday in [1, 4]:
                            self.screen_agents(agent_type,
                            self.Testing.preventive_screening_test_type,\
                             'preventive')
                    # (c.3) testing every 2 days = testing on Mo, Wed & Fri
                    elif interval == 2 and self.weekday in [1, 3, 5]:
                            self.screen_agents(agent_type,
                            self.Testing.preventive_screening_test_type,\
                             'preventive')
                    # No interval specified = no testing, even if testing
                    # mode == preventive
                    elif interval == None:
                        pass
                    else:
                        if self.verbosity > 0:
                            print('not initiating {} preventive screen (wrong weekday)'\
                                    .format(agent_type))
            else:
                # do nothing
                pass

            for agent_type in self.agent_types:
                if not (self.screened_agents['reactive'][agent_type] or \
                        self.screened_agents['follow_up'][agent_type] or \
                        self.screened_agents['preventive'][agent_type]):
                        self.days_since_last_agent_screen[agent_type] += 1


        if self.verbosity > 0: print('* agent interaction *')
        self.datacollector.collect(self)
        self.schedule.step()
        self.Nstep += 1



from scseirx.agent_SEIRX import agent_SEIRX

class family_member(agent_SEIRX):
    '''
    A family member with an infection status
    '''

    def __init__(self, unique_id, unit, model, 
        exposure_duration, time_until_symptoms, infection_duration,
        voluntary_testing, vaccinated, verbosity):
        
        self.type = 'family_member'

        super().__init__(unique_id, unit, model, 
            exposure_duration, time_until_symptoms, infection_duration,
            voluntary_testing, vaccinated, verbosity)


    def step(self):
        '''
        Infection step: if a family member is infected and not in quarantine,
        it interacts with other family members trough the specified 
        contact network and can pass a potential infection.
        Infections are staged here and only applied in the 
        "advance"-step to simulate "simultaneous" interaction
        '''

        # check for external infection in continuous index case modes
        if self.model.index_case in ['continuous'] and \
           self.index_probability > 0:
            self.introduce_external_infection()

        # simulate contacts to other agents if the agent is
        # infected and not in quarantine. Randomly transmit the infection 
        # according to the transmission risk
        if self.infectious:
            if not self.quarantined:
                
                # get contacts to other agent groups according to the
                # interaction network
                family_members = self.get_contacts('family_member')
                teachers = self.get_contacts('teacher')
                students = self.get_contacts('student')

                # code transmission to other agent groups
                # separately to allow for differences in transmission risk
                self.transmit_infection(family_members)
                self.transmit_infection(teachers)
                self.transmit_infection(students)


from scseirx.agent_SEIRX import agent_SEIRX

class employee(agent_SEIRX):
    '''
    An employee with an infection status
    '''

    def __init__(self, unique_id, unit, model, 
        exposure_duration, time_until_symptoms, infection_duration,
        voluntary_testing, vaccinated, verbosity):

        self.type = 'employee'

        super().__init__(unique_id, unit, model, 
            exposure_duration, time_until_symptoms, infection_duration,
            voluntary_testing, vaccinated, verbosity)
        

    def step(self):
        '''
        Infection step: if a employee is infected and not in quarantine, it 
        interacts with other residents and employees trough the specified 

        contact network and can pass a potential infection.
        Infections are staged here and only applied in the 
        "advance"-step to simulate "simultaneous" interaction
        '''

        # check for external infection in continuous index case modes
        if self.model.index_case in ['continuous'] and \
           self.index_probability > 0:
            self.introduce_external_infection()

        # simulate contacts to other agents if the agent is
        # infected and not in quarantine. Randomly transmit the infection 
        # according to the transmission risk
        if self.infectious:
            if not self.quarantined:

                # get contacts to other agent groups according to the
                # interaction network
                residents = self.get_contacts('resident')
                employees = self.get_contacts('employee')

                # code transmission to other agent groups
                # separately to allow for differences in transmission risk
                self.transmit_infection(residents)
                self.transmit_infection(employees)


from setuptools import find_packages, setup
import pathlib

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name='scseirx',
    version='1.4.1',
    packages=find_packages(where='src'),
    package_dir={"": "src"},
    package_data={'scseirx':['data/nursing_home/*.bz2', 'data/school/*.bz2', 'img/*.png']},
    description='A simulation tool to explore the spread of COVID-19 in small communities such as nursing homes or schools via agent-based modeling (ABM) and the impact of prevention measures.',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/JanaLasser/agent_based_COVID_SEIRX",
    author='Jana Lasser',
    author_email='lasser@tugraz.at',
    license='MIT',
    classifiers=[
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.8"],
    install_requires=[
        'numpy>=1.19.2',
        'scipy>=1.6.1',
        'matplotlib==3.3.4',
        'networkx>=2.5',
        'mesa>=0.8.8.1',
        'pandas>=1.2.3'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==6.2.2'],
    test_suite='tests'
)



import numpy as np
import networkx as nx
from mesa import Model
from mesa.time import RandomActivation, SimultaneousActivation
from mesa.datacollection import DataCollector

from scseirx.model_SEIRX import *


## data collection functions ##
def count_S_resident(model):
    S = np.asarray([1 for a in model.schedule.agents if a.type == 'resident' and\
                    a.exposed == False and a.recovered == False \
                    and a.infectious == False]).sum()
    return S
    
    
def count_E_resident(model):
    E = np.asarray(
        [a.exposed for a in model.schedule.agents if a.type == 'resident']).sum()
    return E


def count_I_resident(model):
    I = np.asarray(
        [a.infectious for a in model.schedule.agents if a.type == 'resident']).sum()
    return I


def count_I_symptomatic_resident(model):
    I = np.asarray([a.infectious for a in model.schedule.agents if
        (a.type == 'resident'and a.symptomatic_course)]).sum()
    return I

def count_V_resident(model):
    V = np.asarray([a.vaccinated for a in model.schedule.agents if
        (a.type == 'resident')]).sum()
    return V

def count_I_asymptomatic_resident(model):
    I = np.asarray([a.infectious for a in model.schedule.agents if
        (a.type == 'resident'and a.symptomatic_course == False)]).sum()
    return I


def count_R_resident(model):
    R = np.asarray(
        [a.recovered for a in model.schedule.agents if a.type == 'resident']).sum()
    return R


def count_X_resident(model):
    X = np.asarray(
        [a.quarantined for a in model.schedule.agents if a.type == 'resident']).sum()
    return X


def count_S_employee(model):
    S = np.asarray([1 for a in model.schedule.agents if a.type == 'employee' and\
                    a.exposed == False and a.recovered == False \
                    and a.infectious == False]).sum()
    return S


def count_E_employee(model):
    E = np.asarray(
        [a.exposed for a in model.schedule.agents if a.type == 'employee']).sum()
    return E


def count_I_employee(model):
    I = np.asarray(
        [a.infectious for a in model.schedule.agents if a.type == 'employee']).sum()
    return I


def count_I_symptomatic_employee(model):
    I = np.asarray([a.infectious for a in model.schedule.agents if
        (a.type == 'employee'and a.symptomatic_course)]).sum()
    return I

def count_V_employee(model):
    V = np.asarray([a.vaccinated for a in model.schedule.agents if
        (a.type == 'employee')]).sum()
    return V

def count_I_asymptomatic_employee(model):
    I = np.asarray([a.infectious for a in model.schedule.agents if
        (a.type == 'employee'and a.symptomatic_course == False)]).sum()
    return I


def count_R_employee(model):
    R = np.asarray(
        [a.recovered for a in model.schedule.agents if a.type == 'employee']).sum()
    return R


def count_X_employee(model):
    X = np.asarray(
        [a.quarantined for a in model.schedule.agents if a.type == 'employee']).sum()
    return X


def check_reactive_resident_screen(model):
    return model.screened_agents['reactive']['resident']


def check_follow_up_resident_screen(model):
    return model.screened_agents['follow_up']['resident']


def check_preventive_resident_screen(model):
    return model.screened_agents['preventive']['resident']


def check_reactive_employee_screen(model):
    return model.screened_agents['reactive']['employee']


def check_follow_up_employee_screen(model):
    return model.screened_agents['follow_up']['employee']


def check_preventive_employee_screen(model):
    return model.screened_agents['preventive']['employee']

data_collection_functions = \
    {
    'resident':
        {
        'S':count_S_resident,
        'E':count_E_resident,
        'I':count_I_resident,
        'I_asymptomatic':count_I_asymptomatic_resident,
        'V':count_V_resident,
        'I_symptomatic':count_I_symptomatic_resident,
        'R':count_R_resident,
        'X':count_X_resident
         },
    'employee':
        {
        'S':count_S_employee,
        'E':count_E_employee,
        'I':count_I_employee,
        'I_asymptomatic':count_I_asymptomatic_employee,
        'V':count_V_employee,
        'I_symptomatic':count_I_symptomatic_employee,
        'R':count_R_employee,
        'X':count_X_employee
         }
    }




class SEIRX_nursing_home(SEIRX):


    def __init__(self, G,
        verbosity = 0,
        base_transmission_risk = 0.05,
        testing='diagnostic',
        exposure_duration = [5.0, 1.9],
        time_until_symptoms = [6.4, 0.8],
        infection_duration = [10.91, 3.95],
        quarantine_duration = 10,
        subclinical_modifier = 0.6,
        infection_risk_contact_type_weights = {
            'very_far': 0.1,
            'far': 0.25,
            'intermediate': 0.5,
            'close': 1},
        K1_contact_types = ['close'],
        diagnostic_test_type = 'one_day_PCR',
        preventive_screening_test_type = 'same_day_antigen',
        follow_up_testing_interval = None,
        liberating_testing = False,
        index_case = 'employee',
        agent_types = {
            'employee':      {'screening_interval': None,
                             'index_probability': 0,
                             'mask':False,
                             'vaccination_probability': 0},
            'resident':      {'screening_interval': None,
                             'index_probability': 0,
                             'mask':False,
                             'vaccination_probability': 0}},
        age_transmission_risk_discount = \
             {'slope':-0.02,
              'intercept':1},
        age_symptom_modification = \
             {'slope':0.00777777,
              'intercept':-0.022222},
        mask_filter_efficiency = {'exhale':0, 'inhale':0},
        transmission_risk_ventilation_modifier = 0,
        transmission_risk_vaccination_modifier = {'reception':1, 'transmission':0},
        seed = None):


        super().__init__(G,
            verbosity = verbosity,
            base_transmission_risk = base_transmission_risk,
            testing = testing,
            exposure_duration = exposure_duration,
            time_until_symptoms = time_until_symptoms,
            infection_duration = infection_duration,
            quarantine_duration = quarantine_duration,
            subclinical_modifier = subclinical_modifier,
            infection_risk_contact_type_weights = \
                         infection_risk_contact_type_weights,
            K1_contact_types = K1_contact_types,
            diagnostic_test_type = diagnostic_test_type,
            preventive_screening_test_type = \
                         preventive_screening_test_type,
            follow_up_testing_interval = follow_up_testing_interval,
            liberating_testing = liberating_testing,
            index_case = index_case,
            agent_types = agent_types,
            age_transmission_risk_discount = \
                 age_transmission_risk_discount,
            age_symptom_modification = age_symptom_modification,
            mask_filter_efficiency = mask_filter_efficiency,
            transmission_risk_ventilation_modifier = \
                         transmission_risk_ventilation_modifier,
            transmission_risk_vaccination_modifier = \
                         transmission_risk_vaccination_modifier,
            seed = seed)



        # agent types that are included in preventive, background & follow-up
        # screens
        self.screening_agents = ['employee', 'resident']

        # define, whether or not a multigraph that defines separate connections
        # for every day of the week is used
        self.dynamic_connections = False


        # data collectors to save population counts and agent states every
        # time step
        model_reporters = {}
        for agent_type in self.agent_types:

            for state in ['S','E','I','I_asymptomatic','I_symptomatic','R','X', 'V']:

                model_reporters.update({'{}_{}'.format(state, agent_type):\
                    data_collection_functions[agent_type][state]})

        model_reporters.update(\
            {
            'screen_residents_reactive':check_reactive_resident_screen,
            'screen_residents_follow_up':check_follow_up_resident_screen,
            'screen_residents_preventive':check_preventive_resident_screen,
            'screen_employees_reactive':check_reactive_employee_screen,
            'screen_employees_follow_up':check_follow_up_employee_screen,
            'screen_employees_preventive':check_preventive_employee_screen,
            'N_diagnostic_tests':get_N_diagnostic_tests,
            'N_preventive_screening_tests':get_N_preventive_screening_tests,
            'undetected_infections':get_undetected_infections,
            'predetected_infections':get_predetected_infections,
            'pending_test_infections':get_pending_test_infections
            })

        agent_reporters =\
            {
            'infection_state':get_infection_state,
            'quarantine_state':get_quarantine_state
             }

        self.datacollector = DataCollector(
            model_reporters = model_reporters,
            agent_reporters = agent_reporters)

    def calculate_transmission_probability(self, source, target, base_risk):
        """
        Calculates the risk of transmitting an infection between a source agent
        and a target agent given the model's and agent's properties and the base
        transmission risk.

        Transmission is an independent Bernoulli trial with a probability of
        success p. The probability of transmission without any modifications
        by for example masks or ventilation is given by the base_risk, which
        is calibrated in the model. The probability is modified by contact type
        q1 (also calibrated in the model), infection progression q2
        (from literature), reduction of the viral load due to a sublclinical
        course of the disease q3 (from literature), reduction of exhaled viral
        load of the source by mask wearing q4 (from literature), reduction of
        inhaled viral load by the target q5 (from literature), and ventilation
        of the rooms q6 (from literature).

        Parameters
        ----------
        source : agent_SEIRX
            Source agent that transmits the infection to the target.
        target: agent_SEIRX
            Target agent that (potentially) receives the infection from the
            source.
        base_risk : float
            Probability p of infection transmission without any modifications
            through prevention measures.

        Returns
        -------
        p : float
            Modified transmission risk.
        """
        n1 = source.ID
        n2 = target.ID
        link_type = self.G.get_edge_data(n1, n2)['link_type']

        q1 = self.get_transmission_risk_contact_type_modifier(source, target)
        q2 = self.get_transmission_risk_progression_modifier(source)
        q3 = self.get_transmission_risk_subclinical_modifier(source)
        q9 = self.get_transmission_risk_vaccination_modifier_reception(target)
        q10 = self.get_transmission_risk_vaccination_modifier_transmission(source)

        # contact types where masks and ventilation are irrelevant
        if link_type in ['resident_resident_room', 'resident_resident_table']:
            p = 1 - (1 - base_risk * (1- q1) * (1 - q2) * (1 - q3) * (1 - q9) \
                * (1 - q10))

        # contact types were masks and ventilation are relevant
        elif link_type in ['resident_resident_quarters',
                           'employee_resident_care',
                           'employee_employee_short']:

            q4 = self.get_transmission_risk_exhale_modifier(source)
            q5 = self.get_transmission_risk_inhale_modifier(target)
            q6 = self.get_transmission_risk_ventilation_modifier()

            p = 1 - (1 - base_risk * (1- q1) * (1 - q2) * (1 - q3) * \
                (1 - q4) * (1 - q5) * (1 - q6) * (1 - q9) * (1 - q10))

        else:
            print('unknown link type: {}'.format(link_type))
            p = None
        return p



from mesa import Agent


class agent_SEIRX(Agent):
    '''
    An agent with an infection status. NOTe: this agent is not
    functional on it's own, as it does not implement a step()
    function. Therefore, every agent class that inherits from this
    generic agent class needs to implement their own step() function
    '''

    def __init__(self, unique_id, unit, model,
        exposure_duration, time_until_symptoms, infection_duration, vaccinated,
        voluntary_testing, verbosity):
        super().__init__(unique_id, model)
        self.verbose = verbosity
        self.ID = unique_id
        self.unit = unit
        self.voluntary_testing = voluntary_testing

        ## epidemiological parameters drawn from distributions
        # NOTE: all durations are inclusive, i.e. comparison are "<=" and ">="
        # number of days agents stay infectuous

        # days after transmission until agent becomes infectuous
        self.exposure_duration = exposure_duration
        # days after becoming infectuous until showing symptoms
        self.time_until_symptoms = time_until_symptoms
        # number of days agents stay infectuous
        self.infection_duration = infection_duration
        # vaccinated true or false (depending on chosen probability)
        self.vaccinated = vaccinated


        ## agent-group wide parameters that are stored in the model class
        self.index_probability = self.model.index_probabilities[self.type]
        self.mask = self.model.masks[self.type]

        # try to set the agent's age from information in the graph. The age info
        # is later needed to adjust transmission risk and the probability to
        # develop a symptomatic course. If the age information doees not exist,
        # age is set to 0, the transmission risk remains unmodified and the
        # probability to develop a symptomatic course is set to the intercept
        # specified at model setup
        try: 
            self.age = model.G.nodes(data=True)[self.unique_id]['age']
        except KeyError:
            self.age = 0

        ## age adjustments
        # adjust symptom probability based on age
        self.symptom_probability = \
                        self.age * self.model.age_symptom_modification['slope'] + \
                        self.model.age_symptom_modification['intercept']


        ## infection states
        self.exposed = False
        self.infectious = False
        self.symptomatic_course = False

        if not self.vaccinated and \
            self.model.random.random() <= self.symptom_probability:
            self.symptomatic_course = True


        self.symptoms = False
        self.recovered = False
        self.tested = False
        self.pending_test = False
        self.known_positive = False
        self.quarantined = False

        # sample given for test
        self.sample = None

        # staging states
        self.contact_to_infected = False

        # counters
        self.days_since_exposure = 0
        self.days_quarantined = 0
        self.days_since_tested = 0
        self.transmissions = 0
        self.transmission_targets = {}



    ### generic helper functions that are inherited by other agent classes

    def get_contacts(self, agent_group):
        contacts = [a for a in self.model.schedule.agents if
            (a.type == agent_group and self.model.G.has_edge(self.ID, a.ID))]
        return contacts


    def introduce_external_infection(self):
        if (self.infectious == False) and (self.exposed == False) and\
           (self.recovered == False):
            index_transmission = self.model.random.random()
            if index_transmission <= self.index_probability:
                self.contact_to_infected = True
                if self.verbose > 0:
                    print('{} {} is index case'.format(
                        self.type, self.unique_id))


    def transmit_infection(self, contacts):
        # the basic transmission risk is that between two members of the 
        # same household and has been calibrated to reproduce empirical 
        # household secondary attack rates.
        base_risk = self.model.base_transmission_risk

        for target in contacts:
            if (target.exposed == False) and (target.infectious == False) and \
               (target.recovered == False) and (target.contact_to_infected == False):

                # determine if a transmission occurrs
                p = self.model.calculate_transmission_probability(\
                                self, target, base_risk)
                transmission = self.model.random.random()

                if self.verbose > 1:
                    print('target: {} {}, p: {}'\
                        .format(target.type, target.ID, p))

                if transmission < p:
                    target.contact_to_infected = True
                    self.transmissions += 1

                    # track the state of the agent pertaining to testing at the
                    # moment of transmission to count how many transmissions
                    # occur in which states
                    if self.tested and self.pending_test and \
                        self.sample == 'positive':
                        self.model.pending_test_infections += 1

                    self.transmission_targets.update({target.ID:self.model.Nstep})

                    if self.verbose > 0:
                        print('transmission: {} {} -> {} {} (p: {})'
                        .format(self.type, self.unique_id, \
                                target.type, target.unique_id, p))


    def act_on_test_result(self):
        '''
        Function that gets called by the infection dynamics model class if a
        test result for an agent is returned. The function sets agent states
        according to the result of the test (positive or negative). Adds agents
        with positive tests to the newly_positive_agents list that will be
        used to trace and quarantine close (K1) contacts of these agents. Resets
        the days_since_tested counter and the sample as well as the 
        pending_test flag
        '''

        # the type of the test used in the test for which the result is pending
        # is stored in the pending_test variable
        test_type = self.pending_test

        if self.sample == 'positive':

            # true positive
            if self.model.Testing.tests[test_type]['sensitivity'] >= self.model.random.random():
                self.model.newly_positive_agents.append(self)
                self.known_positive = True

                if self.model.verbosity > 1:
                    print('{} {} returned a positive test (true positive)'
                        .format(self.type, self.ID))

                if self.quarantined == False:
                    self.quarantined = True
                    if self.model.verbosity > 0:
                        print('quarantined {} {}'.format(self.type, self.ID))

            # false negative
            else:
                if self.model.verbosity > 1:
                    print('{} {} returned a negative test (false negative)'\
                        .format(self.type, self.ID))
                self.known_positive = False
                self.model.false_negative += 1

                if self.model.Testing.liberating_testing:
                    self.quarantined = False
                    if self.model.verbosity > 0:
                        print('{} {} left quarantine prematurely'\
                        .format(self.type, self.ID))

            self.days_since_tested = 0
            self.pending_test = False
            self.sample = None

        elif self.sample == 'negative':

            # false positive
            if self.model.Testing.tests[test_type]['specificity'] <= self.model.random.random():
                self.model.newly_positive_agents.append(self)
                self.known_positive = True

                if self.model.verbosity > 1:
                    print('{} {} returned a positive test (false positive)'\
                        .format(self.type, self.ID))

                if self.quarantined == False:
                    self.quarantined = True
                    if self.model.verbosity > 0:
                        print('quarantined {} {}'.format(self.type, self.ID))

            # true negative
            else:
                if self.model.verbosity > 1:
                    print('{} {} returned a negative test (true negative)'\
                        .format(self.type, self.ID))
                self.known_positive = False

                if self.model.Testing.liberating_testing:
                    self.quarantined = False
                    if self.model.verbosity > 0:
                        print('{} {} left quarantine prematurely'\
                        .format(self.type, self.ID))

            self.days_since_tested = 0
            self.pending_test = False
            self.sample = None

    def become_exposed(self):
        if self.verbose > 0:
            print('{} exposed: {}'.format(self.type, self.unique_id))
        self.exposed = True
        self.contact_to_infected = False


    def become_infected(self):
        self.exposed = False
        self.infectious = True

        if self.verbose > 0:
            if self.symptomatic_course:
                print('{} infectious: {} (symptomatic course)'\
                    .format(self.type, self.unique_id))
            else:
                print('{} infectious: {} (asymptomatic course)'\
                    .format(self.type, self.unique_id))


    def show_symptoms(self):
        # determine if agent shows symptoms
        if self.symptomatic_course:
            self.symptoms = True
            if self.model.verbosity > 0:
                print('{} {} shows symptoms'.format(self.type, self.ID))


    def recover(self):
        self.infectious = False
        self.symptoms = False
        self.recovered = True
        self.days_since_exposure = self.infection_duration + 1
        if self.verbose > 0:
            print('{} recovered: {}'.format(self.type, self.unique_id))


    def leave_quarantine(self):
        if self.verbose > 0:
            print('{} released from quarantine: {}'.format(
                self.type, self.unique_id))
        self.quarantined = False
        self.days_quarantined = 0


    def advance(self):
        '''
        Advancing step: applies infections, checks counters and sets infection 
        states accordingly
        '''

        # determine if a transmission to the agent occurred
        if self.contact_to_infected == True:
            self.become_exposed()

        # determine if agent has transitioned from exposed to infected
        if self.days_since_exposure == self.exposure_duration:
            self.become_infected()

        if self.days_since_exposure == self.time_until_symptoms:
            self.show_symptoms()

        if self.days_since_exposure == self.infection_duration:
            self.recover()

        # determine if agent is released from quarantine
        if self.days_quarantined == self.model.quarantine_duration:
            self.leave_quarantine()

        # if there is a pending test result, increase the days the agent has
        # waited for the result by 1 (NOTE: results are collected by the 
        # infection dynamics model class according to days passed since the test)
        if self.pending_test:
            self.days_since_tested += 1

        if self.quarantined:
            self.days_quarantined += 1
            self.model.quarantine_counters[self.type] += 1

        if self.exposed or self.infectious:
            self.days_since_exposure += 1

        # reset tested flag at the end of the agent step
        self.tested = False
        
